from mopidy import backend

from . import translator


class InternetArchivePlaybackProvider(backend.PlaybackProvider):
    def translate_uri(self, uri):
        identifier, filename, _ = translator.parse_uri(uri)
        return self.backend.client.geturl(identifier, filename)



import collections
import re

from mopidy.config import types

import pytest
from mopidy_internetarchive import ConfigMap


def test_deserialize():
    expected = collections.OrderedDict([("a", "1"), ("b", "2"), ("c", "3")])
    type = ConfigMap()
    assert type.deserialize("a|1, b|2 , c|3") == expected
    assert type.deserialize("a|1\n b|2 \nc|3") == expected

    with pytest.raises(ValueError):
        type.deserialize("")
    with pytest.raises(ValueError):
        type.deserialize("a")
    with pytest.raises(ValueError):
        type.deserialize("a|")
    with pytest.raises(ValueError):
        type.deserialize("|1")


def test_deserialize_choice_keys():
    expected = collections.OrderedDict([("a", "1"), ("b", "2"), ("c", "3")])
    type = ConfigMap(keys=types.String(choices=["a", "b", "c"]))
    assert type.deserialize("a|1, b|2 , c|3") == expected
    assert type.deserialize("a|1\n b|2 \nc|3") == expected

    with pytest.raises(ValueError):
        type.deserialize("x|0")


def test_deserialize_integer_keys():
    expected = collections.OrderedDict([(1, "a"), (2, "b"), (3, "c")])
    type = ConfigMap(keys=types.Integer())
    assert type.deserialize("1|a, 2|b , 3|c") == expected
    assert type.deserialize("1|a\n 2|b \n3|c") == expected


def test_deserialize_list_keys():
    expected = collections.OrderedDict([(("a",), "1"), (("b", "c"), "2")])
    type = ConfigMap(keys=types.List())
    assert type.deserialize("a|1\n b, c|2") == expected


def test_deserialize_choice_values():
    expected = collections.OrderedDict([("a", "1"), ("b", "2"), ("c", "3")])
    type = ConfigMap(values=types.String(choices=["1", "2", "3"]))
    assert type.deserialize("a|1, b|2 , c|3") == expected
    assert type.deserialize("a|1\n b|2 \nc|3") == expected

    with pytest.raises(ValueError):
        type.deserialize("x|0")


def test_deserialize_integer_values():
    expected = collections.OrderedDict([("a", 1), ("b", 2), ("c", 3)])
    type = ConfigMap(values=types.Integer())
    assert type.deserialize("a|1, b|2 , c|3") == expected
    assert type.deserialize("a|1\n b|2 \nc|3") == expected

    with pytest.raises(ValueError):
        type.deserialize("x|y")


def test_deserialize_list_values():
    expected = collections.OrderedDict([("a", ("1",)), ("b", ("2", "3"))])
    type = ConfigMap(values=types.List())
    assert type.deserialize("a|1\n b|2,3") == expected


def test_deserialize_optional_values():
    expected = collections.OrderedDict([("a", "1"), ("b", None), ("c", None)])
    type = ConfigMap(values=types.String(optional=True))
    assert type.deserialize("a|1, b| , c") == expected
    assert type.deserialize("a|1\n b| \nc") == expected


def test_delim():
    expected = collections.OrderedDict([("a", "1"), ("b", "2"), ("c", "3")])
    type = ConfigMap(delim=":")
    assert type.deserialize("a:1, b: 2 , c :3") == expected
    assert type.deserialize("a:1\n b: 2 \nc :3") == expected


def test_optional():
    expected = collections.OrderedDict()
    assert ConfigMap(optional=True).deserialize("") == expected

    with pytest.raises(ValueError):
        ConfigMap(optional=False).deserialize("")
    with pytest.raises(ValueError):
        ConfigMap().deserialize("")


def test_serialize():
    type = ConfigMap(keys=types.String(), values=types.Integer())
    value = collections.OrderedDict([("a", 1), ("b", 2), ("c", 3)])
    result = type.serialize(value)
    assert re.match(r"\s*a|1\n\s*b|2\n\s*c|3", result)


def test_serialize_none():
    type = ConfigMap(keys=types.String(), values=types.Integer())
    result = type.serialize(None)
    assert result == ""



import collections

from unittest import mock

import mopidy_internetarchive as ext
import pytest


@pytest.fixture
def config():
    return {
        "internetarchive": {
            "base_url": "http://archive.org",
            "collections": ("audio", "etree", "foo"),
            "audio_formats": ("Flac", "VBR MP3"),
            "image_formats": ("JPEG", "PNG"),
            "browse_limit": None,
            "browse_views": collections.OrderedDict(
                [("title asc", "Title"), ("creator asc", "Creator")]
            ),
            "search_limit": None,
            "search_order": None,
            "cache_size": None,
            "cache_ttl": None,
            "retries": 0,
            "timeout": None,
        },
        "proxy": {},
    }


@pytest.fixture
def root_collections():
    from mopidy.models import Ref

    return [
        Ref.directory(name="Audio Archive", uri="internetarchive:audio"),
        Ref.directory(name="Live Music Archive", uri="internetarchive:etree"),
    ]


@pytest.fixture
def audio_mock():
    audio_mock = mock.Mock()
    return audio_mock


@pytest.fixture
def client_mock():
    client_mock = mock.Mock(spec=ext.client.InternetArchiveClient)
    client_mock.SearchResult = ext.client.InternetArchiveClient.SearchResult
    client_mock.cache = mock.Mock(spec=dict)
    client_mock.search.return_value = client_mock.SearchResult(
        {
            "responseHeader": {"params": {"q": "album"}},
            "response": {
                "numFound": 2,
                "docs": [
                    {
                        "identifier": "etree",
                        "title": "Live Music Archive",
                        "mediatype": "collection",
                    },
                    {
                        "identifier": "audio",
                        "title": "Audio Archive",
                        "mediatype": "collection",
                    },
                ],
            },
        }
    )
    return client_mock


@pytest.fixture
def backend_mock(client_mock, config):
    backend_mock = mock.Mock(spec=ext.backend.InternetArchiveBackend)
    backend_mock.client = client_mock
    return backend_mock


@pytest.fixture
def library(backend_mock, config):
    return ext.library.InternetArchiveLibraryProvider(
        config["internetarchive"], backend_mock
    )


@pytest.fixture
def playback(audio_mock, backend_mock):
    return ext.playback.InternetArchivePlaybackProvider(
        audio_mock, backend_mock
    )



from collections.abc import Sequence

import operator
import urllib.parse

import requests

import cachetools

BASE_URL = "http://archive.org/"


def _session(base_url, retries):
    # TODO: backoff?
    session = requests.Session()
    adapter = requests.adapters.HTTPAdapter(max_retries=retries)
    session.mount(base_url, adapter)
    return session


class InternetArchiveClient:

    pykka_traversable = True

    def __init__(self, base_url=BASE_URL, retries=0, timeout=None):
        self.__base_url = base_url
        self.__session = _session(base_url, retries)
        self.__timeout = timeout
        self.cache = None  # public

    @property
    def proxies(self):
        return self.__session.proxies

    @property
    def useragent(self):
        return self.__session.headers.get("User-Agent")

    @useragent.setter
    def useragent(self, value):
        self.__session.headers["User-Agent"] = value

    @cachetools.cachedmethod(operator.attrgetter("cache"))
    def getitem(self, identifier):
        obj = self.__get("/metadata/%s" % identifier).json()
        if not obj:
            raise LookupError(identifier)
        elif "error" in obj:
            raise LookupError(obj["error"])
        elif "result" in obj:
            return obj["result"]
        else:
            return obj

    def geturl(self, identifier, filename=None):
        if filename:
            path = f"/download/{identifier}/{filename}"
        else:
            path = "/download/%s" % identifier
        return urllib.parse.urljoin(self.__base_url, path)

    def search(self, query, fields=None, sort=None, rows=None, start=None):
        response = self.__get(
            "/advancedsearch.php",
            params={
                "q": query,
                "fl[]": fields,
                "sort[]": sort,
                "rows": rows,
                "start": start,
                "output": "json",
            },
        )
        if response.content:
            return self.SearchResult(response.json())
        else:
            raise self.SearchError(response.url)

    def __get(self, path, params=None):
        return self.__session.get(
            urllib.parse.urljoin(self.__base_url, path),
            params=params,
            timeout=self.__timeout,
        )

    class SearchResult(Sequence):
        def __init__(self, result):
            response = result["response"]
            self.docs = response.get("docs", [])
            self.rowcount = response.get("numFound", None)
            # query is optional, and responseHeader likely to change
            try:
                self.query = result["responseHeader"]["params"]["query"]
            except LookupError:
                self.query = None

        def __getitem__(self, key):
            return self.docs[key]

        def __len__(self):
            return len(self.docs)

        def __iter__(self):
            return iter(self.docs)

    class SearchError(Exception):
        pass


if __name__ == "__main__":
    import argparse
    import logging
    import json
    import sys

    parser = argparse.ArgumentParser()
    parser.add_argument("arg", metavar="PATH | USER | QUERY")
    parser.add_argument("-B", "--base-url", default="http://archive.org")
    parser.add_argument("-f", "--fields", nargs="+")
    parser.add_argument("-i", "--indent", type=int, default=2)
    parser.add_argument("-q", "--query", action="store_true")
    parser.add_argument("-r", "--rows", type=int)
    parser.add_argument("-R", "--retries", type=int, default=0)
    parser.add_argument("-s", "--sort", nargs="+")
    parser.add_argument("-t", "--timeout", type=float)
    parser.add_argument("-v", "--verbose", action="store_true")
    args = parser.parse_args()

    logging.basicConfig(level=logging.DEBUG if args.verbose else logging.WARN)

    client = InternetArchiveClient(args.base_url, args.retries, args.timeout)
    if args.query:
        result = client.search(args.arg, args.fields, args.sort, args.rows)
    else:
        result = client.getitem(args.arg)
    json.dump(result, sys.stdout, default=vars, indent=args.indent)
    sys.stdout.write("\n")



import collections
import logging

from mopidy import backend, models

from . import Extension, translator

logger = logging.getLogger(__name__)


class InternetArchiveLibraryProvider(backend.LibraryProvider):

    root_directory = models.Ref.directory(
        uri=translator.uri(""), name="Internet Archive"
    )

    def __init__(self, config, backend):
        super().__init__(backend)
        self.__collections = config["collections"]
        self.__audio_formats = config["audio_formats"]
        self.__image_formats = config["image_formats"]

        self.__browse_filter = "(mediatype:collection OR format:(%s))" % (
            " OR ".join(map(translator.quote, config["audio_formats"]))
        )
        self.__browse_limit = config["browse_limit"]
        self.__browse_views = config["browse_views"]

        self.__search_filter = "format:(%s)" % (
            " OR ".join(map(translator.quote, config["audio_formats"]))
        )
        self.__search_limit = config["search_limit"]
        self.__search_order = config["search_order"]

        self.__directories = collections.OrderedDict()
        self.__lookup = {}  # track cache for faster lookup

    def browse(self, uri):
        identifier, filename, query = translator.parse_uri(uri)
        if filename:
            return []
        elif identifier and query:
            return self.__browse_collection(identifier, **query)
        elif identifier:
            return self.__browse_item(identifier)
        else:
            return self.__browse_root()

    def get_images(self, uris):
        # map uris to item identifiers
        urimap = collections.defaultdict(list)
        for uri in uris:
            identifier, _, _ = translator.parse_uri(uri)
            if identifier:
                urimap[identifier].append(uri)
            else:
                logger.debug("Not retrieving images for %s", uri)
        # retrieve item images and map back to uris
        results = {}
        for identifier, uris in urimap.items():
            try:
                item = self.backend.client.getitem(identifier)
            except Exception as e:
                logger.error("Error retrieving images for %s: %s", uris, e)
            else:
                results.update(dict.fromkeys(uris, self.__images(item)))
        return results

    def lookup(self, uri):
        try:
            return [self.__lookup[uri]]
        except KeyError:
            logger.debug("Lookup cache miss for %r", uri)
        identifier, filename, _ = translator.parse_uri(uri)
        if identifier:
            tracks = self.__tracks(self.backend.client.getitem(identifier))
            self.__lookup = trackmap = {t.uri: t for t in tracks}
            return [trackmap[uri]] if filename else tracks
        else:
            return []

    def refresh(self, uri=None):
        client = self.backend.client
        if client.cache:
            client.cache.clear()
        self.__directories.clear()
        self.__lookup.clear()

    def search(self, query=None, uris=None, exact=False):
        # sanitize uris
        uris = set(uris or [self.root_directory.uri])
        if self.root_directory.uri in uris:
            uris.update(translator.uri(c) for c in self.__collections)
            uris.remove(self.root_directory.uri)
        # translate query
        try:
            qs = translator.query(query, uris, exact)
        except ValueError as e:
            logger.info("Not searching %s: %s", Extension.dist_name, e)
            return None
        else:
            logger.debug("Internet Archive query: %s" % qs)
        # fetch results
        result = self.backend.client.search(
            f"{qs} AND {self.__search_filter}",
            fields=["identifier", "title", "creator", "date"],
            rows=self.__search_limit,
            sort=self.__search_order,
        )
        logger.debug("Internet Archive result: %s" % list(result))
        return models.SearchResult(
            uri=translator.uri(q=result.query),
            albums=[translator.album(item) for item in result],
        )

    def __browse_collection(self, identifier, sort=("downloads desc",)):
        return [
            translator.ref(res)
            for res in self.backend.client.search(
                f"collection:{identifier} AND {self.__browse_filter}",
                fields=["identifier", "mediatype", "title", "creator"],
                rows=self.__browse_limit,
                sort=sort,
            )
        ]

    def __browse_item(self, identifier):
        if identifier in self.__directories:
            return self.__views(identifier)
        item = self.backend.client.getitem(identifier)
        if item["metadata"]["mediatype"] == "collection":
            return self.__views(identifier)
        tracks = self.__tracks(item)
        self.__lookup = {t.uri: t for t in tracks}  # cache tracks
        return [models.Ref.track(uri=t.uri, name=t.name) for t in tracks]

    def __browse_root(self):
        if not self.__directories:
            result = self.backend.client.search(
                "mediatype:collection AND identifier:(%s)"
                % (" OR ".join(self.__collections)),
                fields=["identifier", "mediatype", "title"],
            )
            objs = {obj["identifier"]: obj for obj in result}
            for identifier in self.__collections:
                try:
                    obj = objs[identifier]
                except KeyError as e:
                    logger.warning("Collection not found: %s", e)
                else:
                    self.__directories[identifier] = translator.ref(obj)
        return list(self.__directories.values())

    def __images(self, item):
        uri = self.backend.client.geturl  # get download URL for images
        return translator.images(item, self.__image_formats, uri)

    def __tracks(self, item, key=lambda t: (t.track_no or 0, t.uri)):
        tracks = translator.tracks(item, self.__audio_formats)
        tracks.sort(key=key)
        return tracks

    def __views(self, identifier):
        refs = []
        for order, name in self.__browse_views.items():
            uri = translator.uri(identifier, sort=order)
            refs.append(models.Ref.directory(name=name, uri=uri))
        return refs



import collections
import pathlib

import pkg_resources

from mopidy import config, ext

__version__ = pkg_resources.get_distribution("Mopidy-InternetArchive").version

SORT_FIELDS = [
    f"{f} {o}"
    for o in ("asc", "desc")
    for f in (
        "addeddate",
        "avg_rating",
        "call_number",
        "createdate",
        "creatorSorter",
        "date",
        "downloads",
        "foldoutcount",
        "headerImage",
        "identifier",
        "imagecount",
        "indexdate",
        "languageSorter",
        "licenseurl",
        "month",
        "nav_order",
        "num_reviews",
        "publicdate",
        "reviewdate",
        "stars",
        "titleSorter",
        "week",
        "year",
    )
]


class ConfigMap(config.ConfigValue):

    default_keys = config.String()

    default_values = config.String()

    def __init__(
        self,
        keys=default_keys,
        values=default_values,
        delim="|",
        optional=False,
    ):
        self.__keys = keys
        self.__values = values
        self.__delim = delim
        self.__optional = optional

    def deserialize(self, value):
        dict = collections.OrderedDict()
        for s in config.List(optional=self.__optional).deserialize(value):
            parts = s.partition(self.__delim)
            key = self.__keys.deserialize(parts[0])
            val = self.__values.deserialize(parts[2])
            dict[key] = val
        return dict

    def serialize(self, value, display=False):
        if not value:
            return ""
        d = config.String().serialize(self.__delim)
        return config.List().serialize(
            [
                self.__keys.serialize(k) + d + self.__values.serialize(v)
                for k, v in value.items()
            ]
        )


class Extension(ext.Extension):

    dist_name = "Mopidy-InternetArchive"
    ext_name = "internetarchive"
    version = __version__

    def get_default_config(self):
        return config.read(pathlib.Path(__file__).parent / "ext.conf")

    def get_config_schema(self):
        schema = super().get_config_schema()
        schema.update(
            base_url=config.String(),
            collections=config.List(),
            audio_formats=config.List(),
            image_formats=config.List(),
            browse_limit=config.Integer(minimum=1, optional=True),
            browse_views=ConfigMap(keys=config.String(choices=SORT_FIELDS)),
            search_limit=config.Integer(minimum=1, optional=True),
            search_order=config.String(choices=SORT_FIELDS, optional=True),
            cache_size=config.Integer(minimum=1, optional=True),
            cache_ttl=config.Integer(minimum=0, optional=True),
            retries=config.Integer(minimum=0),
            timeout=config.Integer(minimum=0, optional=True),
            # no longer used
            browse_order=config.Deprecated(),
            exclude_collections=config.Deprecated(),
            exclude_mediatypes=config.Deprecated(),
            username=config.Deprecated(),
        )
        return schema

    def setup(self, registry):
        from .backend import InternetArchiveBackend

        registry.add("backend", InternetArchiveBackend)



import pykka
from mopidy import backend, httpclient

import cachetools

from . import Extension
from .client import InternetArchiveClient
from .library import InternetArchiveLibraryProvider
from .playback import InternetArchivePlaybackProvider


def _cache(cache_size=None, cache_ttl=None, **kwargs):
    if cache_size is None:
        return None
    elif cache_ttl is None:
        return cachetools.LRUCache(cache_size)
    else:
        return cachetools.TTLCache(cache_size, cache_ttl)


class InternetArchiveBackend(pykka.ThreadingActor, backend.Backend):

    uri_schemes = [Extension.ext_name]

    def __init__(self, config, audio):
        super().__init__()
        ext_config = config[Extension.ext_name]

        self.client = client = InternetArchiveClient(
            ext_config["base_url"],
            retries=ext_config["retries"],
            timeout=ext_config["timeout"],
        )
        product = f"{Extension.dist_name}/{Extension.version}"
        client.useragent = httpclient.format_user_agent(product)
        proxy = httpclient.format_proxy(config["proxy"])
        client.proxies.update({"http": proxy, "https": proxy})
        client.cache = _cache(**ext_config)

        self.library = InternetArchiveLibraryProvider(ext_config, self)
        self.playback = InternetArchivePlaybackProvider(audio, self)



from mopidy import models

import pytest

ITEM = {
    "files": [
        {
            "name": "track02.mp3",
            "title": "Track #2",
            "format": "VBR MP3",
            "track": "02",
        },
        {
            "name": "track01.mp3",
            "title": "Track #1",
            "format": "VBR MP3",
            "track": "01",
        },
    ],
    "metadata": {"identifier": "album", "title": "Album", "mediatype": "audio"},
}

ALBUM = models.Album(name="Album", uri="internetarchive:album")

TRACK1 = models.Track(
    album=ALBUM,
    name="Track #1",
    track_no=1,
    uri="internetarchive:album#track01.mp3",
)

TRACK2 = models.Track(
    album=ALBUM,
    name="Track #2",
    track_no=2,
    uri="internetarchive:album#track02.mp3",
)


def test_lookup_root(library, client_mock):
    assert library.lookup("internetarchive:") == []


def test_lookup_album(library, client_mock):
    client_mock.getitem.return_value = ITEM
    results = library.lookup("internetarchive:album")
    client_mock.getitem.assert_called_once_with("album")
    assert results == [TRACK1, TRACK2]


def test_lookup_track(library, client_mock):
    client_mock.getitem.return_value = ITEM
    results = library.lookup("internetarchive:album#track01.mp3")
    client_mock.getitem.assert_called_once_with("album")
    assert results == [TRACK1]
    # assert lookup cache is used
    client_mock.reset_mock()
    results = library.lookup("internetarchive:album#track02.mp3")
    client_mock.getitem.assert_not_called()
    assert results == [TRACK2]


def test_lookup_refresh(library, client_mock):
    client_mock.getitem.return_value = ITEM
    results = library.lookup("internetarchive:album#track01.mp3")
    client_mock.getitem.assert_called_once_with("album")
    assert results == [TRACK1]
    # clear lookup cache
    library.refresh()
    assert client_mock.cache.clear.called
    # assert lookup cache is cleared
    client_mock.reset_mock()
    results = library.lookup("internetarchive:album#track02.mp3")
    client_mock.getitem.assert_called_once_with("album")
    assert results == [TRACK2]


def test_lookup_unknown(library, client_mock):
    client_mock.getitem.side_effect = LookupError("null")
    with pytest.raises(LookupError):
        library.lookup("internetarchive:null")
    client_mock.getitem.assert_called_once_with("null")



def test_translate_url(playback, client_mock):
    url = "http://archive.org/download/item/file.mp3"
    client_mock.geturl.return_value = url
    result = playback.translate_uri("internetarchive:item#file.mp3")
    assert client_mock.geturl.called_once()
    assert client_mock.geturl.call_args == (("item", "file.mp3"),)
    assert result == url



from mopidy import models

URL = "http://archive.org/download/album/cover.jpg"

ITEM = {
    "files": [
        {
            "name": "track02.mp3",
            "title": "Track #2",
            "format": "VBR MP3",
            "track": "02",
        },
        {
            "name": "track01.mp3",
            "title": "Track #1",
            "format": "VBR MP3",
            "track": "01",
        },
        {"name": "cover.jpg", "format": "JPEG"},
    ],
    "metadata": {"identifier": "album", "title": "Album", "mediatype": "audio"},
}

IMAGES = [models.Image(uri=URL)]


def test_root_images(library, client_mock):
    results = library.get_images(["internetarchive:"])
    client_mock.getitem.assert_not_called()
    assert results == {}


def test_album_images(library, client_mock):
    client_mock.getitem.return_value = ITEM
    client_mock.geturl.return_value = URL
    results = library.get_images(["internetarchive:album"])
    client_mock.getitem.assert_called_once_with("album")
    assert results == {"internetarchive:album": IMAGES}


def test_track_images(library, client_mock):
    client_mock.getitem.return_value = ITEM
    client_mock.geturl.return_value = URL
    results = library.get_images(
        [
            "internetarchive:album#track01.jpg",
            "internetarchive:album#track02.jpg",
        ]
    )
    client_mock.getitem.assert_called_once_with("album")
    assert results == {
        "internetarchive:album#track01.jpg": IMAGES,
        "internetarchive:album#track02.jpg": IMAGES,
    }



from unittest import mock

from mopidy_internetarchive import Extension, backend


def test_get_default_config():
    ext = Extension()

    config = ext.get_default_config()

    assert "[internetarchive]" in config
    assert "enabled = true" in config


def test_get_config_schema():
    ext = Extension()

    schema = ext.get_config_schema()

    assert "audio_formats" in schema
    assert "base_url" in schema
    assert "browse_limit" in schema
    assert "browse_order" in schema
    assert "cache_size" in schema
    assert "cache_ttl" in schema
    assert "collections" in schema
    assert "exclude_collections" in schema
    assert "exclude_mediatypes" in schema
    assert "image_formats" in schema
    assert "retries" in schema
    assert "search_limit" in schema
    assert "search_order" in schema
    assert "timeout" in schema


def test_setup():
    registry = mock.Mock()

    ext = Extension()
    ext.setup(registry)

    registry.add.assert_called_with("backend", backend.InternetArchiveBackend)



from mopidy import models

import pytest
from mopidy_internetarchive import translator


def test_parse_bitrate():
    assert translator.parse_bitrate(None) is None
    assert translator.parse_bitrate("") is None
    assert translator.parse_bitrate([]) is None
    assert translator.parse_bitrate("tmp") is None

    assert translator.parse_bitrate(None, 42) == 42
    assert translator.parse_bitrate("", 42) == 42
    assert translator.parse_bitrate([], 42) == 42
    assert translator.parse_bitrate("tmp", 42) == 42

    assert translator.parse_bitrate("0") == 0
    assert translator.parse_bitrate("1") == 1000
    assert translator.parse_bitrate("42.123") == 42123


def test_parse_date():
    assert translator.parse_date(None) is None
    assert translator.parse_date("") is None
    assert translator.parse_date([]) is None
    assert translator.parse_date("tmp") is None

    assert translator.parse_date(None, 42) == 42
    assert translator.parse_date("", 42) == 42
    assert translator.parse_date([], 42) == 42
    assert translator.parse_date("tmp", 42) == 42

    assert translator.parse_date("2014-02-21T06:00:00Z") == "2014-02-21"
    assert translator.parse_date("2014-02-21") == "2014-02-21"
    assert translator.parse_date("2014-02") == "2014-02-01"
    assert translator.parse_date("2014") == "2014-01-01"


def test_parse_length():
    assert translator.parse_length(None) is None
    assert translator.parse_length("") is None
    assert translator.parse_length([]) is None
    assert translator.parse_length("tmp") is None

    assert translator.parse_length(None, 42) == 42
    assert translator.parse_length("", 42) == 42
    assert translator.parse_length([], 42) == 42
    assert translator.parse_length("tmp", 42) == 42

    assert translator.parse_length("0") == 0
    assert translator.parse_length("1") == 1 * 1000
    assert translator.parse_length("60") == 60 * 1000
    assert translator.parse_length("1:00") == 60 * 1000
    assert translator.parse_length("60:00") == 3600 * 1000
    assert translator.parse_length("1:00:00") == 3600 * 1000


def test_parse_mtime():
    assert translator.parse_mtime(None) is None
    assert translator.parse_mtime("") is None
    assert translator.parse_mtime([]) is None
    assert translator.parse_mtime("tmp") is None

    assert translator.parse_mtime(None, 42) == 42
    assert translator.parse_mtime("", 42) == 42
    assert translator.parse_mtime([], 42) == 42
    assert translator.parse_mtime("tmp", 42) == 42

    assert translator.parse_mtime("0") == 0
    assert translator.parse_mtime("1") == 1


def test_parse_track():
    assert translator.parse_track(None) is None
    assert translator.parse_track("") is None
    assert translator.parse_track([]) is None
    assert translator.parse_track("tmp") is None

    assert translator.parse_track(None, 42) == 42
    assert translator.parse_track("", 42) == 42
    assert translator.parse_track([], 42) == 42
    assert translator.parse_track("tmp", 42) == 42

    assert translator.parse_track("0") == 0
    assert translator.parse_track("1") == 1
    assert translator.parse_track("1/100") == 1
    assert translator.parse_track("1/tmp") == 1


def test_uri(uri=translator.uri):
    assert "internetarchive:" == uri()
    assert "internetarchive:item" == uri("item")
    assert "internetarchive:item#file.mp3" == uri("item", "file.mp3")
    assert "internetarchive:item#file%201.mp3" == uri("item", "file 1.mp3")
    assert "internetarchive:?q=foo" == uri(q="foo")
    assert "internetarchive:?q=foo%20AND%20bar" == uri(q="foo AND bar")
    assert "internetarchive:item?sort=title%20asc" == uri(
        "item", sort="title asc"
    )


def test_ref(ref=translator.ref):
    assert models.Ref.album(name="foo", uri="internetarchive:foo") == ref(
        {"identifier": "foo", "mediatype": "audio"}
    )
    assert models.Ref.directory(name="foo", uri="internetarchive:foo") == ref(
        {"identifier": "foo", "mediatype": "collection"}
    )
    assert models.Ref.directory(name="foo", uri="internetarchive:foo") == ref(
        {"identifier": "foo", "mediatype": "collection"}
    )
    assert models.Ref.album(name="Foo", uri="internetarchive:foo") == ref(
        {"identifier": "foo", "mediatype": "audio", "title": "Foo"}
    )
    assert models.Ref.directory(name="Foo", uri="internetarchive:foo") == ref(
        {"identifier": "foo", "mediatype": "collection", "title": "Foo"}
    )
    # as of Dec 2016, items in oldtimeradio collections return title as list
    assert models.Ref.album(name="#1", uri="internetarchive:otr") == ref(
        {"identifier": "otr", "mediatype": "audio", "title": ["#1", "#2"]}
    )
    assert models.Ref.directory(name="#1", uri="internetarchive:otr") == ref(
        {"identifier": "otr", "mediatype": "collection", "title": ["#1", "#2"]}
    )


def test_artists(artists=translator.artists):
    assert artists({}) is None
    assert artists({"artist": ""}) is None
    assert artists({"creator": ""}) is None
    assert artists({"artist": "", "creator": ""}) is None
    assert artists({"creator": []}) is None
    assert artists({"artist": "", "creator": []}) is None
    assert [models.Artist(name="foo")] == artists({"artist": "foo"})
    assert [models.Artist(name="foo")] == artists({"creator": "foo"})
    assert [models.Artist(name="foo")] == artists(
        {"artist": "foo", "creator": "bar"}
    )
    assert [models.Artist(name="foo")] == artists({"creator": ["foo"]})
    assert [models.Artist(name="foo"), models.Artist(name="bar")] == artists(
        {"creator": ["foo", "bar"]}
    )
    assert [models.Artist(name="foo")] == artists(
        {"artist": "foo", "creator": ["bar", "baz"]}
    )


def test_album(album=translator.album):
    model = models.Album(name="foo", uri="internetarchive:foo")
    assert model == album({"identifier": "foo"})
    assert model.replace(name="Foo") == album(
        {"identifier": "foo", "title": "Foo"}
    )
    assert model.replace(artists=[models.Artist(name="bar")]) == album(
        {"identifier": "foo", "creator": "bar"}
    )
    assert model.replace(date="1970-01-01") == album(
        {"identifier": "foo", "date": "1970-01-01"}
    )


def test_query(query=translator.query):
    assert r'"foo"' == query({"any": ["foo"]})
    assert r'"foo \"bar\""' == query({"any": ['foo "bar"']})
    assert r'"foo" AND "bar"' == query({"any": ["foo", "bar"]})
    assert r'title:("foo")' == query({"album": ["foo"]})
    assert r'title:("foo bar")' == query({"album": ["foo bar"]})
    assert r'title:("foo" "bar")' == query({"album": ["foo", "bar"]})
    assert r'creator:("foo")' == query({"albumartist": ["foo"]})
    assert r'creator:("foo" "bar")' == query({"albumartist": ["foo", "bar"]})
    assert r'creator:("foo")' == query({"artist": ["foo"]})
    assert r'creator:("foo" "bar")' == query({"artist": ["foo", "bar"]})
    assert r'date:"1970"' == query({"date": ["1970"]})
    assert r'date:"1970-01-01"' == query({"date": ["1970-01-01"]})


def test_query_exact(query=translator.query):
    with pytest.raises(ValueError):
        query({"any": ["foo"]}, exact=True)


def test_query_unsupported(query=translator.query):
    with pytest.raises(ValueError):
        query({"track_name": ["foo"]})


def test_query_uris(query=translator.query):
    assert r'"foo"' == query({"any": ["foo"]}, [])
    assert r'"foo"' == query({"any": ["foo"]}, [""])
    assert r'"foo"' == query({"any": ["foo"]}, ["internetarchive:"])

    assert r'"foo" AND collection:(bar)' == query(
        {"any": ["foo"]}, ["internetarchive:bar"]
    )
    assert r'"foo" AND collection:(bar OR baz)' == query(
        {"any": ["foo"]}, ["internetarchive:bar", "internetarchive:baz"]
    )

    with pytest.raises(ValueError):
        query({"any": ["foo"]}, ["internetarchive:?foo"])
    with pytest.raises(ValueError):
        query({"any": ["foo"]}, ["internetarchive:#foo"])



from mopidy import models


def test_search_any(library, client_mock):
    client_mock.search.return_value = client_mock.SearchResult(
        {
            "responseHeader": {"params": {"query": "album"}},
            "response": {
                "docs": [
                    {
                        "identifier": "album1",
                        "title": "Album #1",
                        "mediatype": "audio",
                    },
                    {
                        "identifier": "album2",
                        "title": "Album #2",
                        "mediatype": "etree",
                    },
                ],
                "numFound": 2,
            },
        }
    )
    result = library.search(dict(any=["album"]))
    assert client_mock.search.called_once()
    assert result == models.SearchResult(
        uri="internetarchive:?q=album",
        albums=[
            models.Album(name="Album #1", uri="internetarchive:album1"),
            models.Album(name="Album #2", uri="internetarchive:album2"),
        ],
    )


def test_search_unknown(library, client_mock):
    result = library.search(dict(foo=["bar"]))
    client_mock.search.assert_not_called()
    assert result is None



from mopidy import models

COLLECTION = {
    "metadata": {
        "identifier": "directory",
        "title": "Directory",
        "mediatype": "collection",
    }
}

ITEM = {
    "files": [
        {
            "name": "track02.mp3",
            "title": "Track #2",
            "format": "VBR MP3",
            "track": "02",
        },
        {
            "name": "track01.mp3",
            "title": "Track #1",
            "format": "VBR MP3",
            "track": "01",
        },
    ],
    "metadata": {"identifier": "album", "title": "Album", "mediatype": "audio"},
}

TRACKS = [
    models.Ref.track(name="Track #1", uri="internetarchive:album#track01.mp3"),
    models.Ref.track(name="Track #2", uri="internetarchive:album#track02.mp3"),
]

VIEWS = [
    models.Ref.directory(
        name="Title", uri="internetarchive:directory?sort=title%20asc"
    ),
    models.Ref.directory(
        name="Creator", uri="internetarchive:directory?sort=creator%20asc"
    ),
]


def test_has_root_directory(library):
    assert library.root_directory == models.Ref.directory(
        name="Internet Archive", uri="internetarchive:"
    )


def test_browse_root_directory(library, client_mock, root_collections):
    results = library.browse(library.root_directory.uri)
    assert client_mock.search.called_once()
    assert results == root_collections


def test_browse_collection(library, client_mock):
    client_mock.getitem.return_value = COLLECTION
    results = library.browse("internetarchive:directory")
    client_mock.getitem.assert_called_once_with("directory")
    assert results == VIEWS


def test_browse_audio(library, client_mock):
    client_mock.getitem.return_value = ITEM
    results = library.browse("internetarchive:album")
    client_mock.getitem.assert_called_once_with("album")
    assert results == TRACKS


def test_browse_video(library, client_mock):
    client_mock.getitem.return_value = {
        "files": [],
        "metadata": {"identifier": "item", "mediatype": "video"},
    }
    results = library.browse("internetarchive:album")
    client_mock.getitem.assert_called_once_with("album")
    assert results == []


def test_browse_view(library, client_mock):
    client_mock.search.return_value = client_mock.SearchResult(
        {
            "responseHeader": {"params": {"q": "album"}},
            "response": {
                "docs": [
                    {
                        "identifier": "album",
                        "title": "Album",
                        "mediatype": "audio",
                    },
                    {
                        "identifier": "directory",
                        "title": "Directory",
                        "mediatype": "collection",
                    },
                ],
                "numFound": 2,
            },
        }
    )
    results = library.browse("internetarchive:audio?sort=title%20asc")
    assert library.backend.client.search.called_once()
    assert results == [
        models.Ref.album(name="Album", uri="internetarchive:album"),
        models.Ref.directory(name="Directory", uri="internetarchive:directory"),
    ]


def test_browse_file(library, client_mock):
    results = library.browse("internetarchive:album#file.mp3")
    client_mock.getitem.assert_not_called()
    client_mock.search.assert_not_called()
    assert results == []



import collections
import datetime
import logging
import re

import uritools
from mopidy.models import Album, Artist, Image, Ref, Track

from . import Extension

DURATION_RE = re.compile(
    r"""
(?:
    (?:(?P<hours>\d+):)?
    (?P<minutes>\d+):
)?
(?P<seconds>\d+)
""",
    flags=re.VERBOSE,
)

ISODATE_RE = re.compile(
    r"""
(\d{4})
(?:\-(\d{2}))?
(?:\-(\d{2}))?
""",
    flags=re.VERBOSE,
)

QUOTE_RE = re.compile(r'([+!(){}\[\]^"~*?:\\]|\&\&|\|\|)')

_QUERYMAP = {
    "any": lambda values: (" AND ".join(map(quote, values))),
    "album": lambda values: ("title:(%s)" % " ".join(map(quote, values))),
    "albumartist": lambda values: (
        "creator:(%s)" % " ".join(map(quote, values))
    ),
    "artist": lambda values: ("creator:(%s)" % " ".join(map(quote, values))),
    "date": lambda values: (
        # TODO: sanitize, not quote date! date:(2014-01-01) gives error
        " AND ".join("date:%s" % quote(value) for value in values)
    ),
}

logger = logging.getLogger(__name__)


def parse_bitrate(string, default=None):
    if string:
        try:
            return int(float(string) * 1000)
        except Exception:
            logger.warning("Invalid Internet Archive bitrate: %r", string)
    return default


def parse_date(string, default=None):
    if string:
        try:
            return "-".join(ISODATE_RE.match(string).groups("01"))
        except Exception:
            logger.warning("Invalid Internet Archive date: %r", string)
    return default


def parse_length(string, default=None):
    if string:
        try:
            groups = DURATION_RE.match(string).groupdict("0")
            d = datetime.timedelta(**{k: int(v) for k, v in groups.items()})
            return int(d.total_seconds() * 1000)
        except Exception:
            logger.warning("Invalid Internet Archive length: %r", string)
    return default


def parse_mtime(string, default=None):
    if string:
        try:
            return int(string)
        except Exception:
            logger.warning("Invalid Internet Archive mtime: %r", string)
    return default


def parse_track(string, default=None):
    if string:
        try:
            return int(string.partition("/")[0])
        except Exception:
            logger.warning("Invalid Internet Archive track: %r", string)
    return default


def parse_uri(uri):
    parts = uritools.urisplit(uri)
    return parts.path, parts.getfragment(), parts.getquerydict()


def uri(identifier="", filename=None, scheme=Extension.ext_name, **kwargs):
    if filename:
        return uritools.uricompose(scheme, path=identifier, fragment=filename)
    elif kwargs:
        return uritools.uricompose(scheme, path=identifier, query=kwargs)
    else:
        return f"{scheme}:{identifier}"


def name(obj):
    title = obj.get("title")
    if isinstance(title, str):
        return title
    elif hasattr(title, "__getitem__"):
        return title[0]
    else:
        return obj.get("identifier")


def ref(obj, uri=uri):
    identifier = obj["identifier"]
    mediatype = obj["mediatype"]
    if mediatype == "collection":
        return Ref.directory(name=name(obj), uri=uri(identifier))
    else:
        return Ref.album(name=name(obj), uri=uri(identifier))


def artists(obj):
    artist = obj.get("artist", obj.get("creator"))
    if not artist:
        return None
    elif isinstance(artist, str):
        return [Artist(name=artist)]
    else:
        return [Artist(name=name) for name in artist]


def album(obj, uri=uri):
    return Album(
        uri=uri(obj["identifier"]),
        name=name(obj),
        artists=artists(obj),
        date=parse_date(obj.get("date")),
    )


def files(item, formats):
    byname = {}
    byformat = collections.defaultdict(list)
    for obj in item["files"]:
        byname[obj["name"]] = obj
        byformat[obj["format"]].append(obj)
    try:
        files = next(byformat[f] for f in formats if f in byformat)
    except StopIteration:
        return []
    else:
        return [dict(byname.get(f.get("original"), {}), **f) for f in files]


def images(item, formats, uri=uri):
    identifier = item["metadata"]["identifier"]
    images = []
    for obj in files(item, formats):
        images.append(Image(uri=uri(identifier, obj["name"])))
    return images


def tracks(item, formats, uri=uri):
    identifier = item["metadata"]["identifier"]
    track = Track(album=album(item["metadata"]))
    tracks = []
    for obj in files(item, formats):
        filename = obj.get("name")
        tracks.append(
            track.replace(
                uri=uri(identifier, filename),
                name=obj.get("title", filename),
                artists=artists(obj) or track.album.artists,
                genre=obj.get("genre"),
                track_no=parse_track(obj.get("track")),
                length=parse_length(obj.get("length")),
                bitrate=parse_bitrate(obj.get("bitrate")),
                last_modified=parse_mtime(obj.get("mtime")),
            )
        )
    return tracks


def query(query, uris=None, exact=False):
    if exact:
        raise ValueError("Exact queries not supported")
    terms = []
    for key, values in query.items() if query else []:
        try:
            term = _QUERYMAP[key](values)
        except KeyError:
            raise ValueError('Keyword "%s" not supported' % key)
        else:
            terms.append(term)
    collections = []
    for uri in uris or []:
        parts = uritools.urisplit(uri)
        if parts.path:
            collections.append(parts.path)
        elif not parts.query and not parts.fragment:
            pass  # root URI?
        else:
            raise ValueError('Cannot search "%s"' % uri)
    if collections:
        terms.append("collection:(%s)" % " OR ".join(collections))
    return " AND ".join(terms)


def quote(value, re=QUOTE_RE):
    return '"%s"' % re.sub(r"\\\1", value)
